"""This module provides local environment describers."""

from __future__ import annotations

import itertools
import logging
import re
import subprocess
from shutil import which

import numpy as np
import pandas as pd
from monty.io import zopen
from monty.tempfile import ScratchDir
from pymatgen.core import Composition, Element, Molecule, Structure
from pymatgen.core.periodic_table import get_el_sp

from maml.base import BaseDescriber, describer_type
from maml.utils import pool_from, to_composition

__all__ = [
    "BPSymmetryFunctions",
    "BispectrumCoefficients",
    "SiteElementProperty",
    "SmoothOverlapAtomicPosition",
]


logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


@describer_type("site")
class BispectrumCoefficients(BaseDescriber):
    r"""
    Bispectrum coefficients to describe the local environment of each atom.
    Lammps is required to perform this computation.

    Reference:
    @article{bartok2010gaussian,
             title={Gaussian approximation potentials: The
                accuracy of quantum mechanics, without the electrons},
             author={Bart{\'o}k, Albert P and Payne, Mike C
                and Kondor, Risi and Cs{\'a}nyi, G{\'a}bor},
             journal={Physical review letters},
             volume={104}, number={13}, pages={136403}, year={2010}, publisher={APS}}
    """

    def __init__(
        self,
        rcutfac: float,
        twojmax: int,
        element_profile: dict,
        quadratic: bool = False,
        pot_fit: bool = False,
        include_stress: bool = False,
        feature_batch: str = "pandas_concat",
        **kwargs,
    ):
        """
        Args:
            rcutfac (float): The rcutfac used in computing the cutoff between
                elements. true cutoff between i, j = rcutfac * (Ri + Rj),
                where Ri and Rj are cutoff for elements i and j
            twojmax (int): Band limit for bispectrum components.
            element_profile (dict): Parameters (cutoff radius 'r' and weight 'w')
                related to each element, e.g.,
                {'Na': {'r': 4.5, 'w': 0.9},
                 'Cl': {'r': 5.0, 'w': 3.0}}
            quadratic (bool): Whether including quadratic terms.
                Default to False.
            pot_fit (bool): Whether combine the dataframe for potential fitting.
            include_stress (bool): Whether to include stress components.
            feature_batch: way to batch together a list of features
            **kwargs: keyword args to specify memory, verbose, and n_jobs.
        """
        from maml.apps.pes import SpectralNeighborAnalysis

        self.calculator = SpectralNeighborAnalysis(
            rcutfac=rcutfac, twojmax=twojmax, element_profile=element_profile, quadratic=quadratic
        )
        self.rcutfac = rcutfac
        self.twojmax = twojmax
        self.elements = sorted(element_profile.keys(), key=lambda x: Element(x))
        self.element_profile = element_profile
        self.quadratic = quadratic
        self.pot_fit = pot_fit
        self.include_stress = include_stress
        super().__init__(feature_batch=feature_batch, **kwargs)

    @property
    def subscripts(self) -> list:
        """
        The subscripts (2j1, 2j2, 2j) of all bispectrum components
        involved.
        """
        return self.calculator.get_bs_subscripts(self.twojmax)

    def transform_one(self, structure: Structure) -> pd.DataFrame:
        """
        Args:
            structure (Structure): Pymatgen Structure object.
        """
        columns = ["-".join([str(i) for i in s]) for s in self.subscripts]
        if self.quadratic:
            columns += [
                "-".join([f"{i}{j}{k}" for i, j, k in s])
                for s in itertools.combinations_with_replacement(self.subscripts, 2)
            ]

        raw_data = self.calculator.calculate([structure])

        def process(output, combine):
            b, db, vb, e = output
            df = pd.DataFrame(b, columns=columns)
            if combine:
                df_add = pd.DataFrame({"element": e, "n": np.ones(len(e))})
                df_b = df_add.join(df)
                n_atoms = df_b.shape[0]
                b_by_el = [df_b[df_b["element"] == e] for e in self.elements]
                sum_b = [df[df.columns[1:]].sum(axis=0) for df in b_by_el]
                hstack_b = pd.concat(sum_b, keys=self.elements)
                hstack_b = hstack_b.to_frame().T / n_atoms
                hstack_b = hstack_b.fillna(0)
                dbs = np.split(db, len(self.elements), axis=1)
                dbs = np.hstack([np.insert(d.reshape(-1, len(columns)), 0, 0, axis=1) for d in dbs])
                db_index = [f"{i}_{d}" for i in df_b.index for d in "xyz"]
                df_db = pd.DataFrame(dbs, index=db_index, columns=hstack_b.columns)
                if self.include_stress:
                    vbs = np.split(vb.sum(axis=0), len(self.elements))
                    vbs = np.hstack([np.insert(v.reshape(-1, len(columns)), 0, 0, axis=1) for v in vbs])
                    volume = structure.volume
                    vbs = vbs / volume * 160.21766208  # from eV to GPa
                    vb_index = ["xx", "yy", "zz", "yz", "xz", "xy"]
                    df_vb = pd.DataFrame(vbs, index=vb_index, columns=hstack_b.columns)
                    df = pd.concat([hstack_b, df_db, df_vb])
                else:
                    df = pd.concat([hstack_b, df_db])
            return df

        return process(raw_data[0], self.pot_fit)

    @property
    def feature_dim(self) -> int | None:
        """Bispectrum feature dimension."""
        n = 0
        for i in range(self.twojmax + 1):
            for j in range(i + 1):
                for k in range(j - i, min(self.twojmax, i + j) + 1, 2):
                    if k >= i:
                        n += 1
        if self.pot_fit:
            n += 1
        n_elements = len(self.elements)
        n *= n_elements
        return n


@describer_type("site")
class SmoothOverlapAtomicPosition(BaseDescriber):
    r"""
    Smooth overlap of atomic positions (SOAP) to describe the local environment
    of each atom.

    Reference:
    @article{bartok2013representing,
             title={On representing chemical environments},
             author={Bart{\'o}k, Albert P and Kondor, Risi and Cs{\'a}nyi, G{\'a}bor},
             journal={Physical Review B},
             volume={87}, number={18}, pages={184115}, year={2013}, publisher={APS}}
    """

    def __init__(
        self,
        cutoff: float,
        l_max: int = 8,
        n_max: int = 8,
        atom_sigma: float = 0.5,
        feature_batch: str = "pandas_concat",
        **kwargs,
    ):
        """

        Args:
            cutoff (float): Cutoff radius.
            l_max (int): The band limit of spherical harmonics basis function.
                Default to 8.
            n_max (int): The number of radial basis function. Default to 8.
            atom_sigma (float): The width of gaussian atomic density. Default to 0.5.
            feature_batch (str): way to batch together a list of features
            **kwargs: keyword args to specify memory, verbose, and n_jobs.
        """
        from maml.apps.pes import GAPotential

        self.operator = GAPotential()
        self.cutoff = cutoff
        self.l_max = l_max
        self.n_max = n_max
        self.atom_sigma = atom_sigma
        super().__init__(feature_batch=feature_batch, **kwargs)

    def transform_one(self, structure: Structure) -> pd.DataFrame:
        """
        Args:
            structure (Structure): Pymatgen Structure object.
        """
        if not which("quip"):
            raise RuntimeError(
                "quip has not been found.\n", "Please refer to https://github.com/libAtoms/QUIP for ", "further detail."
            )

        atoms_filename = "structure.xyz"

        exe_command = ["quip"]
        exe_command.append(f"atoms_filename={atoms_filename}")

        descriptor_command = ["soap"]
        descriptor_command.append(f"cutoff={self.cutoff}")
        descriptor_command.append(f"l_max={self.l_max}")
        descriptor_command.append(f"n_max={self.n_max}")
        descriptor_command.append(f"atom_sigma={self.atom_sigma}")

        atomic_numbers = [str(element.number) for element in sorted(np.unique(structure.species))]
        n_Z = len(atomic_numbers)
        n_species = len(atomic_numbers)
        Z = "{" + " ".join(atomic_numbers) + "}"
        species_Z = "{" + " ".join(atomic_numbers) + "}"
        descriptor_command.append(f"n_Z={n_Z}")
        descriptor_command.append(f"Z={Z}")
        descriptor_command.append(f"n_species={n_species}")
        descriptor_command.append(f"species_Z={species_Z}")

        exe_command.append("descriptor_str={" + " ".join(descriptor_command) + "}")

        with ScratchDir("."):
            _ = self.operator.write_cfgs(filename=atoms_filename, cfg_pool=pool_from([structure]))
            descriptor_output = "output"
            with open(descriptor_output, "w") as f, subprocess.Popen(exe_command, stdout=f) as p:
                stdout = p.communicate()[0]
                rc = p.returncode
            if rc != 0:
                error_msg = f"quip/soap exited with return code {rc}"
                msg = stdout.decode("utf-8").split("\n")[:-1]
                try:
                    error_line = next(i for i, m in enumerate(msg) if m.startswith("ERROR"))
                    error_msg += ", ".join(msg[error_line:])
                except Exception:
                    error_msg += msg[-1]
                raise RuntimeError(error_msg)

            with zopen(descriptor_output, "rt") as f:  # type: ignore
                lines = f.read()

            descriptor_pattern = re.compile("DESC(.*?)\n", re.S)
            return pd.DataFrame([np.array(c.split(), dtype=np.float64) for c in descriptor_pattern.findall(lines)])


@describer_type("site")
class BPSymmetryFunctions(BaseDescriber):
    r"""
    Behler-Parrinello symmetry function to describe the local environment
    of each atom.

    Reference:
    @article{behler2007generalized,
            title={Generalized neural-network representation of
                high-dimensional potential-energy surfaces},
            author={Behler, J{\"o}rg and Parrinello, Michele},
            journal={Physical review letters},
            volume={98},
            number={14},
            pages={146401},
            year={2007},
            publisher={APS}}
    """

    def __init__(
        self,
        cutoff: float,
        r_etas: np.ndarray,
        r_shift: np.ndarray,
        a_etas: np.ndarray,
        zetas: np.ndarray,
        lambdas: np.ndarray,
        feature_batch: str = "pandas_concat",
        **kwargs,
    ):
        """
        Args:
            cutoff (float): The cutoff distance.
            r_etas (numpy.ndarray): η in radial function.
            r_shift (numpy.ndarray): Rs in radial function.
            a_etas (numpy.ndarray): η in angular function.
            zetas (numpy.ndarray): ζ in angular function.
            lambdas (numpy.ndarray): λ in angular function. Default to (1, -1).
            feature_batch: str = 'pandas_concat',
            **kwargs: keyword args to specify memory, verbose, and n_jobs.
        """
        self.cutoff = cutoff
        self.r_etas = np.array(r_etas)[None, :, None]
        self.r_shift = np.array(r_shift)[None, None, :]
        self.a_etas = np.array(a_etas)[None, :, None, None]
        self.zetas = np.array(zetas)[None, None, :, None]
        self.lambdas = np.array(lambdas)[None, None, None, :]
        super().__init__(feature_batch=feature_batch, **kwargs)

    def transform_one(self, structure: Structure) -> pd.DataFrame:
        """
        Args:
            structure (Structure): Pymatgen Structure object.
        """
        elements = sorted(structure.symbol_set, key=lambda sym: get_el_sp(sym).X)
        all_neighbors = structure.get_all_neighbors(self.cutoff)
        data = []

        for atom_idx, neighbors in enumerate(all_neighbors):
            center = structure[atom_idx].coords
            site_symmfuncs: list[np.ndarray] = []
            sorted_neighbors = sorted(neighbors, key=lambda neighbor: neighbor.species_string)
            temp_dict = {
                element: [(neighbor.coords, neighbor.nn_distance) for neighbor in group]
                for element, group in itertools.groupby(sorted_neighbors, key=lambda neighbor: neighbor.species_string)
            }
            for specie in elements:
                distances = np.array([nn_distance for _, nn_distance in temp_dict[specie]])[:, None, None]
                g2 = np.sum(
                    np.exp(-self.r_etas * (distances - self.r_shift) ** 2) * self._fc(distances), axis=0  # type: ignore
                )
                site_symmfuncs.extend(g2.ravel().tolist())

            for specie1, specie2 in itertools.combinations_with_replacement(elements, 2):
                group1, group2 = temp_dict[specie1], temp_dict[specie2]

                c = (
                    itertools.combinations(range(len(group1)), 2)
                    if specie1 == specie2
                    else itertools.product(range(len(group1)), range(len(group2)))
                )
                index_combination = np.array(list(c)).T.tolist()

                coords_group1 = np.array([coords for coords, _ in group1])
                coords_group2 = np.array([coords for coords, _ in group2])
                distances_group1 = np.array([nn_distance for _, nn_distance in group1])
                distances_group2 = np.array([nn_distance for _, nn_distance in group2])

                v1 = coords_group1[index_combination[0]]
                v2 = coords_group2[index_combination[1]]
                d1 = distances_group1[index_combination[0]]
                d2 = distances_group2[index_combination[1]]
                d3 = np.linalg.norm(v1 - v2, axis=1)
                cosines = np.sum((v1 - center) * (v2 - center), axis=1) / (d1 * d2)
                cosines = cosines[:, None, None, None]
                distances = np.stack((d1, d2, d3))
                cutoffs = np.prod(self._fc(distances), axis=0)  # type: ignore
                cutoffs = np.atleast_1d(cutoffs)[:, None, None, None]
                powers = np.sum(distances**2, axis=0)[:, None, None, None]
                g4 = np.sum(
                    (1 + self.lambdas * cosines) ** self.zetas
                    * np.exp(-self.a_etas * powers)
                    * cutoffs
                    * 2.0 ** (1 - self.zetas),
                    axis=0,
                )
                site_symmfuncs.extend(g4.ravel().tolist())

            data.append(site_symmfuncs)
        return pd.DataFrame(data)

    def _fc(self, r: float) -> np.ndarray:
        """
        Cutoff function to decay the symmetry functions at vicinity of radial cutoff.

        Args:
            r (float): The pair distance.
        """
        return 0.5 * (np.cos(np.pi * r / self.cutoff) + 1) * (r <= self.cutoff)


@describer_type("site")
class SiteElementProperty(BaseDescriber):
    """
    Site specie property describers. For a structure or composition, return
    an unordered set of site specie properties.
    """

    def __init__(self, feature_dict: dict | None = None, output_weights: bool = False, **kwargs):
        """
        Args:
            feature_dict (dict): mapping from atomic number of feature vectors
            output_weights (bool): whether to output the site fraction,
                used in disordered compositions.
            **kwargs: passthrough.
        """
        self.feature_dict = feature_dict
        self.output_weights = output_weights
        super().__init__(feature_batch=None, **kwargs)

    @staticmethod
    def _get_keys(c: Composition) -> tuple[list[int], list[float]]:
        d = {str(i): j for i, j in c._data.items()}
        str_z = {str(i): i.Z for i in c.elements}
        elements = list(d.keys())
        z_values: list[int] = [str_z[i] for i in elements]
        weights: list[float] = [d[i] for i in elements]
        return z_values, weights

    def transform_one(
        self, obj: str | Composition | Structure | Molecule  # type: ignore
    ) -> list[np.ndarray] | np.ndarray:
        """
        Transform one object to features.

        Args:
            obj (str/Composition/Structure/Molecule): object to transform

        Returns:
            features array
        """
        if isinstance(obj, Structure | Molecule):
            keys = []
            weights = []
            for i in obj:
                d = i.species._data
                for k, w in d.items():
                    keys.append(k.Z)
                    weights.append(w)
        else:
            comp = to_composition(obj)
            keys, weights = self._get_keys(comp)

        n = len(keys)

        features = [self.feature_dict[i] for i in keys] if self.feature_dict is not None else keys  # type: ignore
        features = np.reshape(features, (n, -1))  # type: ignore
        weights = np.reshape(weights, (n,))  # type: ignore
        if self.output_weights:
            return [features, weights]  # type: ignore

        int_weights = weights.astype(int)  # type: ignore
        if not np.allclose(int_weights, weights):
            raise ValueError(
                "Number of atoms are not integers and the describers"
                " cannot output single feature matrix. Try set "
                "output_weights = True"
            )
        return np.repeat(features, int_weights, axis=0)

    @property
    def feature_dim(self):
        """Feature dimension."""
        if self.feature_dict is None:
            return None
        key = next(iter(self.feature_dict.keys()))
        return np.array(self.feature_dict[key]).size



"""Selectors."""

from __future__ import annotations

import inspect
from collections import defaultdict
from itertools import combinations
from typing import TYPE_CHECKING

import numpy as np
from joblib import Parallel, delayed
from scipy.linalg import lstsq
from scipy.optimize import NonlinearConstraint, minimize
from sklearn.linear_model import LinearRegression
from sklearn.metrics import get_scorer

if TYPE_CHECKING:
    from collections.abc import Callable


class BaseSelector:
    """
    Feature selector. This is meant to work on relatively smaller
    number of features.
    """

    def __init__(self, coef_thres: float = 1e-6, method: str = "SLSQP"):
        """
        Base selector
        Args:
            coef_thres (float): threshold to discard certain coefficients
            method (str): optimization methods in scipy.optimize.minimize.
        """
        self.coef_thres = coef_thres
        self.is_fitted = False
        self.coef_: np.ndarray | None = None
        self.method = method
        self.indices: np.ndarray | None = None

    def select(self, x: np.ndarray, y: np.ndarray, options: dict | None = None) -> np.ndarray | None:
        """
        Select feature indices from x
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            options (dict): options in the optimizations provided
                to scipy.optimize.minimize
        Returns: list of int indices.
        """
        n_data, n_dim = x.shape
        options = options or {"maxiter": 1e4, "ftol": 1e-12}
        res = minimize(
            lambda beta: self.construct_loss(x=x, y=y, beta=beta),
            [0] * n_dim,
            jac=self.construct_jac(x=x, y=y),
            method=self.method,
            constraints=self.construct_constraints(x=x, y=y),
            options=options,
        )
        if res.status != 0:
            raise RuntimeError(f"Not converged, status {res.status}")
        self.is_fitted = True
        self.coef_ = res.x
        # output coefficient indices that are above certain thresholds
        self.indices = np.where(np.abs(self.coef_) > self.coef_thres)[0]  # type: ignore
        self.coef_[np.where(np.abs(self.coef_) <= self.coef_thres)[0]] = 0.0  # type: ignore
        return self.indices

    def construct_loss(self, x: np.ndarray, y: np.ndarray, beta: np.ndarray) -> float:
        """
        Get loss function from data and tentative coefficients beta
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): N coefficients
        Returns: loss value.
        """
        raise NotImplementedError

    def construct_constraints(
        self, x: np.ndarray, y: np.ndarray, beta: np.ndarray | None = None
    ) -> dict | list | NonlinearConstraint | None:
        """
        Get constraints dictionary from data, e.g.,
        {"func": lambda beta: fun(x, y, beta), "type": "ineq"}.

        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): parameter to optimize
        Returns: dict of constraints.
        """
        return None

    def construct_jac(self, x: np.ndarray, y: np.ndarray) -> Callable | None:
        """
        Jacobian of cost function
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
        Returns: Jacobian function.
        """
        return None

    def evaluate(self, x: np.ndarray, y: np.ndarray, metric: str = "neg_mean_absolute_error") -> float:
        """
        Evaluate the linear models using x, and y test data.

        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            metric (str): scorer function, used with
                sklearn.metrics.get_scorer
        Returns:
        """
        metric_func = get_scorer(metric)
        lr = LinearRegression(fit_intercept=False)
        lr.coef_ = self.coef_[self.indices]  # type: ignore
        lr.intercept_ = 0
        return metric_func(lr, x[:, self.indices], y)

    def get_coef(self) -> np.ndarray | None:
        """
        Get coefficients
        Returns: the coefficients array.
        """
        return self.coef_

    def get_feature_indices(self) -> np.ndarray | None:
        """
        Get selected feature indices.

        Returns: ndarray
        """
        return self.indices

    def predict(self, x: np.ndarray) -> np.ndarray:
        """
        Predict the results using sparsified coefficients.

        Args:
            x (np.ndarray): design matrix
        Returns: ndarray
        """
        return x[:, self.indices].dot(self.coef_[self.indices])  # type: ignore

    def compute_residual(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        """
        Compute.

        Args:
            x (np.ndarray): design matrix
            y (np.ndarray): target vector
        Returns: residual vector.
        """
        return y - self.predict(x)

    @classmethod
    def _get_param_names(cls):
        init = getattr(cls.__init__, "deprecated_original", cls.__init__)
        if init is object.__init__:
            return []
        init_signature = inspect.signature(init)
        parameters = [p for p in init_signature.parameters.values() if p.name != "self" and p.kind != p.VAR_KEYWORD]
        for p in parameters:
            if p.kind == p.VAR_KEYWORD:
                raise RuntimeError(
                    "scikit-learn estimators should always "
                    "specify their parameters in the signature"
                    " of their __init__ (no varargs)."
                    f" {cls} with constructor {init_signature} doesn't "
                    " follow this convention."
                )
        return sorted(p.name for p in parameters)

    def get_params(self):
        """
        Get params for this selector.

        Returns: mapping of string to any
                parameter names mapped to their values

        """
        out = {}
        for key in self._get_param_names():
            value = getattr(self, key, None)
            out[key] = value
        return out

    def set_params(self, **params):
        """
        Set the parameters of this selector
        Args:
            **params: dict
            Selector parameters.

        Returns:
            self: selector instance

        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        valid_params = self.get_params()

        nested_params = defaultdict(dict)  # grouped by prefix
        for key, value in params.items():
            key, delim, sub_key = key.partition("__")
            if key not in valid_params:
                raise ValueError(
                    f"Invalid parameter {key} for selector {self}. "
                    "Check the list of available parameters "
                    "with `estimator.get_params().keys()`."
                )

            if delim:
                nested_params[key][sub_key] = value
            else:
                setattr(self, key, value)
                valid_params[key] = value

        for key, sub_params in nested_params.items():
            valid_params[key].set_params(**sub_params)

        return self


class DantzigSelector(BaseSelector):
    """
    Equation 11 in
    https://orfe.princeton.edu/~jqfan/papers/06/SIS.pdf
    and reference in https://projecteuclid.org/download/pdfview_1/euclid.aos/1201012958.
    """

    def __init__(self, lambd, sigma=1.0, **kwargs):
        """
        Dantzig selector.

        Args:
            lambd: tunable parameter
            sigma: standard deviation of the error.
            **kwargs: passthrough.
        """
        self.lambd = lambd
        self.sigma = sigma
        super().__init__(**kwargs)

    def construct_loss(self, x, y, beta) -> float:
        """
        Get loss function from data and tentative coefficients beta.

        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): N coefficients
        Returns: loss value.
        """
        return np.sum(np.abs(beta)).item()

    def construct_jac(self, x: np.ndarray, y: np.ndarray) -> Callable:
        """
        Jacobian of cost functions.

        Args:
            x: ndarray
            y: ndarray
        Returns: callable
        """

        def _jac(beta):
            sign = np.sign(beta)
            sign[np.abs(sign) < 0.1] = 1.0
            sign *= 30.0  # multiply the gradients to get better convergence
            return sign

        return _jac

    def construct_constraints(
        self, x: np.ndarray, y: np.ndarray, beta: np.ndarray | None = None
    ) -> NonlinearConstraint:
        """
        Get constraints dictionary from data, e.g.,
        {"func": lambda beta: fun(x, y, beta), "type": "ineq"}.

        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): placeholder
        Returns: dict of constraints.
        """

        def _constraint(beta):
            return np.linalg.norm(x.T @ (y - x @ beta), np.infty)

        def _jac(beta):
            vec = x.T @ (y - x @ beta)
            max_ind = np.argmax(np.abs(vec))
            der = np.zeros_like(vec.ravel())
            der[max_ind] = np.sign(vec[max_ind])
            return -x.T.dot(x).dot(der)

        return NonlinearConstraint(_constraint, -np.inf, self.lambd * self.sigma, jac=_jac)


class PenalizedLeastSquares(BaseSelector):
    """
    Penalized least squares. In addition to minimizing the sum of squares loss,
    it adds an additional penalty to the coefficients.
    """

    def construct_loss(self, x: np.ndarray, y: np.ndarray, beta: np.ndarray) -> float:
        """
        Construct the loss function. An extra penalty term is added
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): N coefficients
        Returns: sum of errors.
        """
        n = x.shape[0]
        return 1.0 / (2 * n) * np.sum((y - x.dot(beta)) ** 2) + self.penalty(beta, x=x, y=y)

    def _sse_jac(self, x, y, beta):
        n = x.shape[0]
        return 1.0 / n * (y - x.dot(beta)).T.dot(-x)

    def _penalty_jac(self, x, y, beta):
        return 0.0

    def construct_jac(self, x: np.ndarray, y: np.ndarray):
        """
        Construct the jacobian of loss function
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
        Returns: jacobian vector.
        """

        def _jac(beta):
            return self._sse_jac(x, y, beta) + self._penalty_jac(x, y, beta)

        return _jac

    def construct_constraints(self, x: np.ndarray, y: np.ndarray, beta: np.ndarray | None = None) -> list[dict | None]:
        """
        No constraints
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): placeholder only
        Returns: a list of dictionary constraints.
        """
        return []

    def penalty(self, beta: np.ndarray, x: np.ndarray | None = None, y: np.ndarray | None = None) -> float:
        """
        Calculate the penalty from input x, output y and coefficient beta
        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            beta (np.ndarray): N coefficients
        Returns: penalty value.
        """
        return 0.0


class SCAD(PenalizedLeastSquares):
    """
    Smoothly clipped absolute deviation (SCAD),
    equation 12 and 13 in https://orfe.princeton.edu/~jqfan/papers/06/SIS.pdf.
    """

    def __init__(self, lambd: float | np.ndarray, a: float = 3.7, **kwargs):
        """
        Smoothly clipped absolute deviation.

        Args:
            lambd (float or list of floats): The weights for the penalty
            a (float): hyperparameter in SCAD penalty.
            **kwargs: passthrough
        """
        self.lambd = lambd
        self.a = a
        super().__init__(**kwargs)

    def penalty(self, beta: np.ndarray, x: np.ndarray | None = None, y: np.ndarray | None = None) -> float:
        """
        Calculate the SCAD penalty from input x, output y
            and coefficient beta
        Args:
            beta (np.ndarray): N coefficients
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
        Returns: penalty value.
        """
        beta_abs = np.abs(beta)
        penalty = (
            self.lambd * beta_abs * (beta_abs <= self.lambd)
            + -(beta_abs**2 - 2 * self.a * self.lambd * beta_abs + self.lambd**2)
            / (2 * (self.a - 1))
            * (beta_abs > self.lambd)
            * (beta_abs <= self.a * self.lambd)
            + (self.a + 1) * self.lambd**2 / 2.0 * (beta_abs > self.a * self.lambd)
        )
        return np.sum(penalty).item()

    def _penalty_jac(self, x, y, beta):
        beta = np.abs(beta)
        z = self.a * self.lambd - beta
        z[z < 0] = 0
        return self.lambd * (beta <= self.lambd + z / ((self.a - 1) * self.lambd) * (beta > self.lambd))


class Lasso(PenalizedLeastSquares):
    """Simple Lasso regression."""

    def __init__(self, lambd, **kwargs):
        """
        Lasso regression with lambda * norm_1(beta) as penalty.

        Args:
            lambd (float): weights for the penalty
            **kwargs: passthrough
        """
        self.lambd = lambd
        super().__init__(**kwargs)

    def penalty(self, beta: np.ndarray, x: np.ndarray | None = None, y: np.ndarray | None = None) -> float:
        """
        Calculate the penalty from input x, output y and coefficient beta.

        Args:
            beta (np.ndarray): N coefficients
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
        Returns: penalty value.
        """
        beta_abs = np.abs(beta)
        return np.sum(self.lambd * beta_abs).item()

    def _penalty_jac(self, x, y, beta):
        sign = np.sign(beta)
        sign[np.abs(sign) < 0.2] = 1
        return self.lambd * sign


class AdaptiveLasso(PenalizedLeastSquares):
    """Adaptive lasso regression using OLS coefficients as the root-n estimator coefficients."""

    def __init__(self, lambd, gamma, **kwargs):
        """
        Adaptive lasso regression.

        Args:
            lambd (float or list of floats): lambda values
            gamma (float): exponential for hat(beta)
            **kwargs: passthrough
        """
        self.lambd = lambd
        self.gamma = gamma
        self.w = 1
        super().__init__(**kwargs)

    def select(self, x, y, options=None) -> np.ndarray | None:
        """
        Select feature indices from x.

        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
            options (dict): options in the optimizations provided
                to scipy.optimize.minimize
        Returns: list of int indices.
        """
        self.w = self.get_w(x, y)
        return super().select(x, y, options)

    def get_w(self, x, y) -> np.ndarray:
        """
        Get adaptive weights from data.

        Args:
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
        Returns: coefficients array.
        """
        beta_hat = lstsq(x, y)[0]
        return 1.0 / np.abs(beta_hat) ** self.gamma

    def penalty(self, beta: np.ndarray, x: np.ndarray | None = None, y: np.ndarray | None = None) -> float:
        """
        Calculate the penalty from input x, output y and coefficient beta.

        Args:
            beta (np.ndarray): N coefficients
            x (np.ndarray): MxN input data array
            y (np.ndarray): M output targets
        Returns: penalty value.
        """
        return np.sum(self.lambd * self.w * np.abs(beta)).item()

    def _penalty_jac(self, x, y, beta):
        sign = np.sign(beta)
        sign[np.abs(sign) < 0.2] = 1
        return self.lambd * self.w * sign


class L0BrutalForce(BaseSelector):
    """
    Brutal force combinatorial screening of features.
    This method takes all possible combinations of features
    and optimize the following loss function
        1/2 * mean((y-x @ beta)**2) + lambd * |beta|_0.
    """

    def __init__(self, lambd: float, **kwargs):
        """
        Initialization of L0 optimization.

        Args:
            lambd (float): penalty term
            **kwargs: passthrough
        """
        self.lambd = lambd
        super().__init__(**kwargs)

    def select(self, x: np.ndarray, y: np.ndarray, options: dict | None = None, n_job: int = 1) -> np.ndarray | None:
        """
        L0 combinatorial optimization.

        Args:
            x (np.ndarray): design matrix
            y (np.ndarray): target vector
            options: Dict of options.
            n_job (int): number of cpu
        Returns:
        """
        n, p = x.shape
        index_array = list(range(p))

        def _lstsq(c):
            x_comb = x[:, c]
            beta = lstsq(x_comb, y)[0]
            res = 1.0 / 2 * np.mean((x_comb.dot(beta) - y) ** 2)
            penalty = self.lambd * len(c)
            res += penalty
            return res

        indices = []
        for p_temp in range(1, p + 1):
            indices.extend(list(combinations(index_array, p_temp)))
        loss = Parallel(n_jobs=n_job)(delayed(_lstsq)(comb) for comb in indices)
        argmin = np.argmin(loss)
        self.indices = np.array(indices[argmin])
        x_temp = x[:, self.indices]
        self.coef_ = np.zeros_like(x[0, :])
        self.coef_[self.indices] = lstsq(x_temp, y)[0]  # type: ignore
        return self.indices



"""Dummy test systems."""

from __future__ import annotations

from pymatgen.core import Composition, Lattice, Molecule, Structure

from ._data_conversion import to_array
from ._inspect import get_param_types

DUMMY_OBJECTS = {
    "str": "H2O",
    "composition": Composition("H2O"),
    "structure": Structure(Lattice.cubic(3.167), ["Mo", "Mo"], [[0, 0, 0], [0.5, 0.5, 0.5]]),
    "molecule": Molecule(["C", "O"], [[0, 0, 0], [1, 0, 0]]),
}


def get_describer_dummy_obj(instance):
    """
    For a describers, get a dummy object for transform_one.
    This relies on the type hint.

    Args:
        instance (BaseDescriber): describers instance
    """
    obj_type = getattr(instance, "describer_type", None)
    if obj_type is not None:
        return DUMMY_OBJECTS[obj_type.lower()]
    arg_types = get_param_types(instance.transform_one)
    arg_type = next(iter(arg_types.values()))
    str_t = str(arg_type)
    if "." in str_t:
        str_t = str_t.rsplit(".", maxsplit=1)[-1]
    return DUMMY_OBJECTS[str_t]


def feature_dim_from_test_system(describer):
    """
    Get feature size from a test system.

    Args:
        describer (BaseDescriber): describers instance
    """
    dummy_obj = get_describer_dummy_obj(describer)
    features = to_array(describer.transform_one(dummy_obj))
    if features.ndim == 1:
        return None

    return features.shape[-1]



"""Convert data list to docs or pool existing data lists for training."""

from __future__ import annotations

import numpy as np
import pandas as pd
from pymatgen.core import Structure


def doc_from(structure, energy=None, force=None, stress=None):
    """
    Method to convert structure and its properties into doc
    format for further processing. If properties are None, zeros
    array will be used.

    Args:
        structure (Structure): Pymatgen Structure object.
        energy (float): The total energy of the structure.
        force (np.array): The (m, 3) forces array of the structure
            where m is the number of atoms in structure.
        stress (list/np.array): The (6, ) stresses array of the
            structure.

    Returns:
        (dict)
    """
    energy = energy if energy is not None else 0
    force = force if force is not None else np.zeros((len(structure), 3))
    stress = stress if stress is not None else np.zeros(6)
    outputs = dict(energy=energy, forces=force, virial_stress=stress)
    return dict(structure=structure.as_dict(), num_atoms=len(structure), outputs=outputs)


def pool_from(structures, energies=None, forces=None, stresses=None):
    """
    Method to convert structures and their properties in to
    datapool format.

    Args:
        structures ([Structure]): The list of Pymatgen Structure object.
        energies ([float]): The list of total energies of each structure
            in structures list.
        forces ([np.array]): List of (m, 3) forces array of each structure
            with m atoms in structures list. m can be varied with each
            single structure case.
        stresses (list): List of (6, ) virial stresses of each
            structure in structures list.

    Returns:
        ([dict])
    """
    energies = energies if energies is not None else [None] * len(structures)
    forces = forces if forces is not None else [None] * len(structures)
    stresses = stresses if stresses is not None else [None] * len(structures)
    return [
        doc_from(structure, energy, force, stress)
        for structure, energy, force, stress in zip(structures, energies, forces, stresses, strict=False)
    ]


def convert_docs(docs, include_stress=False, **kwargs):
    """
    Method to convert a list of docs into objects, e.g.,
    Structure and DataFrame.

    Args:
        docs ([dict]): List of docs. Each doc should have the same
            format as one returned from .dft.parse_dir.
        include_stress (bool): Whether to include stress components.
        **kwargs: Passthrough.

    Returns:
        A list of structures, and a DataFrame with energy and force
        data in 'y_orig' column, data type ('energy' or 'force') in
        'dtype' column, No. of atoms in 'n' column sharing the same row
        of energy data while 'n' being 1 for the rows of force data.

    """
    structures, y_orig, n, dtype = [], [], [], []
    for d in docs:
        structure = Structure.from_dict(d["structure"]) if isinstance(d["structure"], dict) else d["structure"]
        outputs = d["outputs"]
        force_arr = np.array(outputs["forces"])
        assert force_arr.shape == (len(structure), 3), "Wrong force array not matching structure"
        force_arr = force_arr.ravel()
        if include_stress:
            virial_stress = outputs["virial_stress"]
            y = np.concatenate(([outputs["energy"]], force_arr, virial_stress))
            dtype.extend(["energy"] + ["force"] * len(force_arr) + ["stress"] * len(virial_stress))
        else:
            y = np.concatenate(([outputs["energy"]], force_arr))
            dtype.extend(["energy"] + ["force"] * len(force_arr))
        y_orig.append(y)
        n.append(np.insert(np.ones(len(y) - 1), 0, d["num_atoms"]))

        structures.append(structure)
    df = pd.DataFrame(dict(y_orig=np.concatenate(y_orig), n=np.concatenate(n), dtype=dtype))
    for k, v in kwargs.items():
        df[k] = v
    return structures, df


def to_array(x):
    """
    Convert x into numerical array
    Args:
        x: x can be a dataframe, a list or an array
    return np.ndarray.
    """
    if isinstance(x, pd.DataFrame):
        return x.to_numpy()
    if isinstance(x, list):
        return np.array([to_array(i) for i in x])
    if isinstance(x, np.ndarray):
        return x
    if isinstance(x, str | int | float):
        return x
    raise ValueError("Not recognized data type")



"""This file contains the PCA implementation."""

from __future__ import annotations

import logging

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.decomposition import PCA

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class PrincipalComponentAnalysis(BaseEstimator, TransformerMixin):
    """Wrap around PCA in scikit-learn to support weighting PCs."""

    def __init__(self, weighting_PCs=True):
        """
        Args:
            weighting_PCs: Whether to weight PCs with explained variances.
        """
        self.pca = PCA()
        self.weighting_PCs = weighting_PCs

    def fit(self, normalized_features):
        """
        Fit PCA with normalized features as input.

        Args:
            normalized_features: An array of normalized features with fixed
                dimensionality. The features have to be normalized first
                beforehand.

        Returns: self
        """
        self.pca.fit(normalized_features)
        return self

    def transform(self, normalized_features):
        """
        Transform normalized features into weighted or unweighted PCA features
        and select the first m PCs according to Kaiser's rule.

        Args:
            normalized_features: An array of normalized features with fixed
                dimensionality. The features have to be normalized first
                beforehand.

        Returns:
            An array of weighted or unweighted PCA feature with dimension of
            N x m, where N and m are the numbers of input normalized features
            and selected PCs according to Kaiser's rule, respectively.
        """
        m = len([e for e in self.pca.explained_variance_ if e > 1])
        explained_variance = self.pca.explained_variance_ratio_
        logger.info(f"Selected first {m} PCs, explaining {100 * sum(explained_variance[:m]):.2f}% variance")
        if not self.weighting_PCs:
            return self.pca.transform(normalized_features)[:, :m]
        return self.pca.transform(normalized_features)[:, :m] * explained_variance[:m]



"""
Implementation of GB energy prediction paper
Weike Ye, Hui Zheng, Chi Chen, Shyue Ping Ong
"A Universal Machine Learning Model for Elemental Grain Boundary Energies"
arXiv preprint arXiv: 2201.11991 (2022).
"""



"""Module implements the BayesianOptimizer."""

from __future__ import annotations

import warnings
from copy import copy
from typing import TYPE_CHECKING

import numpy as np
from pymatgen.core.operations import SymmOp
from pymatgen.core.structure import Lattice, Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from sklearn.base import clone
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RationalQuadratic

from maml.apps.bowsr.acquisition import AcquisitionFunction, ensure_rng, propose_query_point
from maml.apps.bowsr.perturbation import LatticePerturbation, WyckoffPerturbation, get_standardized_structure
from maml.apps.bowsr.preprocessing import DummyScaler, StandardScaler
from maml.apps.bowsr.target_space import TargetSpace

if TYPE_CHECKING:
    from maml.apps.bowsr.model import EnergyModel


def struct2perturbation(
    structure: Structure,
    use_symmetry: bool = True,
    wyc_tol: float = 0.3 * 1e-3,
    abc_tol: float = 1e-3,
    angle_tol: float = 2e-1,
    symprec: float = 1e-2,
) -> tuple[list[WyckoffPerturbation], list[int], dict, LatticePerturbation]:
    """
    Get the symmetry-driven perturbation of the structure.

    Args:
        structure (Structure): Pymatgen Structure object.
        use_symmetry (bool): Whether to use constraint of symmetry to reduce
            parameters space.
        wyc_tol (float): Tolerance for wyckoff symbol determined coordinates.
        abc_tol (float): Tolerance for lattice lengths determined by crystal system.
        angle_tol (float): Tolerance for lattice angles determined by crystal system.
        symprec (float): Tolerance for symmetry finding.

    Return:
        wps (list): WyckoffPerturbations for derivation of symmetrically
            unique sites. Used to derive the coordinates of the sites.
        indices (list): Indices of symmetrically unique sites.
        mapping (dict): A mapping dictionary that maps equivalent atoms
            onto each other.
        lp (LatticePerturbation): LatticePerturbation for derivation of lattice
            of the structure.
    """
    sa = SpacegroupAnalyzer(structure, symprec=symprec)
    sd = sa.get_symmetry_dataset()
    spg_int_number = sd["number"]
    wyckoffs = sd["wyckoffs"]
    symm_ops = [
        SymmOp.from_rotation_and_translation(rotation, translation, tol=0.1)
        for rotation, translation in zip(sd["rotations"], sd["translations"], strict=False)
    ]

    if use_symmetry:
        equivalent_atoms = sd["equivalent_atoms"]
        mapping = dict(zip(range(structure.num_sites), equivalent_atoms, strict=False))
        symmetrically_distinct_indices = np.unique(equivalent_atoms)
        indices = symmetrically_distinct_indices.tolist()
    else:
        indices = list(range(structure.num_sites))
        mapping = dict(zip(indices, indices, strict=False))

    wps = []
    for i in indices:
        wyckoff_symbol = wyckoffs[i]
        wp = WyckoffPerturbation(spg_int_number, wyckoff_symbol, symmetry_ops=symm_ops, use_symmetry=use_symmetry)
        wp.sanity_check(structure[i], wyc_tol=wyc_tol)
        wps.append(wp)
    # wps = [wp for wp in wps]

    lp = LatticePerturbation(spg_int_number, use_symmetry=use_symmetry)
    lp.sanity_check(structure.lattice, abc_tol=abc_tol, angle_tol=angle_tol)

    return (wps, indices, mapping, lp)  # type: ignore


def atoms_crowded(structure: Structure, cutoff_distance: float = 1.1) -> bool:
    """
    Identify whether structure is unreasonable because the atoms are "too close".

    Args:
        structure (Structure): Pymatgen Structure object.
        cutoff_distance (float): The minimum allowed atomic distance.
    """
    distance_matrix = copy(structure.distance_matrix)
    distance_matrix[distance_matrix == 0] = np.inf
    return np.min(distance_matrix) < cutoff_distance


class BayesianOptimizer:
    """Bayesian optimizer used to optimize the structure."""

    def __init__(
        self,
        model: EnergyModel,
        structure: Structure,
        relax_coords: bool = True,
        relax_lattice: bool = True,
        use_symmetry: bool = True,
        use_scaler: bool = True,
        noisy: bool = True,
        seed: int | None = None,
        **kwargs,
    ):
        """
        Args:
            model (EnergyModel): energy model exposes `predict_energy` method.
            structure (Structure): Pymatgen Structure object.
            relax_coords (bool): Whether to relax symmetrical coordinates.
            relax_lattice (bool): Whether to relax lattice.
            use_symmetry (bool): Whether to use constraint of symmetry to reduce
                parameters space.
            use_scaler (bool): Whether to use scaler for Gaussian process regression.
            noisy (bool): Whether to perform noise-based Bayesian optimization
                (predictive noisy test data) or noise-free Bayesian optimization
                (predictive GP posterior).
            seed (int): Seeded rng for random number generator. None
                for an unseeded rng.
            **kwargs: Passthrough.
        """
        random_state = ensure_rng(seed)
        self.model = model
        self.noisy = noisy
        self.use_symmetry = use_symmetry
        scaler = StandardScaler() if use_scaler else DummyScaler()
        self.scaler = scaler

        structure.remove_oxidation_states()
        standardized_structure = get_standardized_structure(structure)
        wps, _, _, lp = struct2perturbation(standardized_structure, **kwargs)

        u = 0
        while not (all(wp.fit_site for wp in wps) and lp.fit_lattice) and u < 3:
            standardized_structure = get_standardized_structure(standardized_structure)
            wps, _, _, lp = struct2perturbation(standardized_structure, use_symmetry, **kwargs)
            u += 1
        if not (all(wp.fit_site for wp in wps) and lp.fit_lattice):
            raise ValueError("Standard structures can not be obtained.")

        self.structure = standardized_structure

        wps, indices, mapping, lp = struct2perturbation(self.structure, use_symmetry, **kwargs)
        wyckoff_dims = [wp.dim for wp in wps]
        abc_dim, angles_dim = lp.dims
        self._space = TargetSpace(
            target_func=self.get_formation_energy,
            wps=wps,
            abc_dim=abc_dim,
            angles_dim=angles_dim,
            relax_coords=relax_coords,
            relax_lattice=relax_lattice,
            scaler=scaler,  # type: ignore
            random_state=random_state,
        )
        self.wps = wps
        self.indices = indices
        self.mapping = mapping
        self.lp = lp
        self.relax_coords = relax_coords
        self.relax_lattice = relax_lattice
        self.wyckoff_dims = wyckoff_dims
        self.abc_dim = abc_dim
        self.angles_dim = angles_dim

        # Initialize internal GaussianProcessRegressor
        self._gpr = GaussianProcessRegressor(kernel=RationalQuadratic(length_scale=1.0), alpha=0.028**2)

    def get_derived_structure(self, x: np.ndarray) -> Structure:
        """
        Get the derived structure.

        Args:
            x (ndarray): The input of getting perturbed structure.
        """
        struct = self.structure.copy()
        lattice = struct.lattice
        species = [dict(site.species.as_dict()) for site in struct]
        frac_coords = struct.frac_coords

        if self.relax_coords:
            x_wyckoff = np.concatenate(
                [
                    self.wps[i].perturbation_mode(x[sum(self.wyckoff_dims[:i]) : sum(self.wyckoff_dims[:i]) + num])
                    for i, num in enumerate(self.wyckoff_dims)
                ]
            )

            for i, (idx, wp) in enumerate(zip(self.indices, self.wps, strict=False)):
                p = wp.standardize(frac_coords[idx]) + x_wyckoff[i * 3 : (i + 1) * 3]
                orbit = wp.get_orbit(p)

                assert len(orbit) == len([k for k, v in self.mapping.items() if v == idx])

                frac_coords[[k for k, v in self.mapping.items() if v == idx]] = orbit

        x_lattice = (
            self.lp.perturbation_mode(x[-(self.abc_dim + self.angles_dim) :]) if self.relax_lattice else np.zeros(6)
        )
        abc = lattice.abc
        angles = lattice.angles
        lattice_parameters = np.array(abc + angles) + x_lattice
        lattice = Lattice.from_parameters(*lattice_parameters)

        return Structure(lattice=lattice, species=species, coords=frac_coords)

    def get_formation_energy(self, x: np.ndarray) -> float:
        """
        Calculate the formation energy of the perturbed structure. Absolute value
        is calculated on practical purpose of maximization of target function in
        Bayesian optimization.

        Args:
            x (ndarray): The input of formation energy calculation.
        """
        derived_struct = self.get_derived_structure(x)
        return -self.model.predict_energy(derived_struct)

    def add_query(self, x: np.ndarray) -> float:
        """
        Add query point into the TargetSpace.

        Args:
            x (ndarray): Query point.
        """
        try:
            target = self.space.probe(x)
        except:  # noqa
            target = 0
        return target

    def propose(self, acquisition_function: AcquisitionFunction, n_warmup: int, sampler: str) -> np.ndarray:
        """
        Suggest the next most promising point.

        Args:
            acquisition_function (AcquisitionFunction): AcquisitionFunction.
            n_warmup (int): Number of randomly sampled points to select the initial
                point for minimization.
            sampler (str): Sampler. Options are Latin Hyperparameter Sampling and uniform sampling.
        """
        # Sklearn's GaussianProcessRegressor throws a large number of warnings at times
        # which are unnecessary to be seen.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.gpr.fit(self.space.params, self.space.target)

        if self.noisy:
            y_max = np.max(self.gpr.predict(self.space.params))
            noise = 0.028
        else:
            y_max = np.max(self.space.target)
            noise = 0.0

        # Find the next point that maximize the acquisition function.
        return propose_query_point(
            acquisition=acquisition_function.calculate,
            scaler=self.scaler,
            gpr=self.gpr,
            y_max=y_max,
            noise=noise,
            bounds=self.space.bounds,
            random_state=self.space.random_state,
            sampler=sampler,
            n_warmup=n_warmup,
        )

    def optimize(
        self,
        n_init: int,
        n_iter: int,
        acq_type: str = "ei",
        kappa: float = 2.576,
        xi: float = 0.0,
        n_warmup: int = 1000,
        is_continue: bool = False,
        sampler: str = "lhs",
        **gpr_params,
    ) -> None:
        """
        Optimize the coordinates and/or lattice of the structure by minimizing the
        model predicted formation energy of the structure. Model prediction error
        can be considered as a constant white noise.

        Args:
            n_init (int): The number of initial points.
            n_iter (int): The number of iteration steps.
            acq_type (str): The type of acquisition function. Three choices are given,
                ucb: Upper confidence bound,
                ei: Expected improvement,
                poi: Probability of improvement.
            kappa (float): Tradeoff parameter used in upper confidence bound formulism.
            xi (float): Tradeoff parameter between exploitation and exploration.
            n_warmup (int): Number of randomly sampled points to select the initial
                point for minimization.
            is_continue (bool): whether to continue previous run without resetting GPR
            sampler (str): Sampler generating initial points. "uniform" or "lhs".
            **gpr_params: Passthrough.
        """
        # print('Optimize begins!')
        self.set_gpr_params(**gpr_params)
        if not is_continue:
            self._gpr = clone(self._gpr)
        acq = AcquisitionFunction(acq_type=acq_type, kappa=kappa, xi=xi)

        # Add the initial structure into known data
        if n_init != 0:
            if sampler == "lhs":
                params = np.concatenate([np.zeros((1, self.space.dim)), self.space.lhs_sample(n_init)])
            else:
                params = np.concatenate(
                    ([np.zeros(self.space.dim)], [self.space.uniform_sample() for _ in range(n_init)])
                )
            self.scaler.fit(params)  # type: ignore

            for x in params:
                self.add_query(x)

        iteration = 0
        while iteration < n_iter:
            x_next = self.propose(acquisition_function=acq, n_warmup=n_warmup, sampler=sampler)
            self.add_query(x_next)
            iteration += 1

    def get_optimized_structure_and_energy(self, cutoff_distance: float = 1.1) -> tuple:
        """
        Args:
            cutoff_distance (float): Cutoff distance of the allowed shortest atomic distance in reasonable structures.
                When the cutoff_distance is 0, any structures will be considered reasonable.
        """
        optimized_structure = self.structure.copy()
        idx = 0
        for idx in np.argsort(self.space.target)[::-1]:
            optimized_structure = self.get_derived_structure(
                self.scaler.inverse_transform(self.space.params[idx])  # type: ignore
            )
            if not atoms_crowded(optimized_structure, cutoff_distance=cutoff_distance):
                break
        return (optimized_structure, -self.space.target[idx])

    def set_space_empty(self) -> None:
        """Empty the target space."""
        self.space.set_empty()

    def set_bounds(self, **bounds_parameter) -> None:
        """Set the bound value of wyckoff perturbation and lattice perturbation."""
        if bounds_parameter.get("abc_bound") is not None:
            abc_bound = bounds_parameter.get("abc_bound")
            bounds_parameter.pop("abc_bound")
        elif bounds_parameter.get("abc_ratio") is not None:
            abc_ratio = bounds_parameter.get("abc_ratio")
            bounds_parameter.pop("abc_ratio")
            abc_bound = np.array(self.lp.abc)[:, np.newaxis] * abc_ratio
        else:
            abc_bound = np.array(self.lp.abc)[:, np.newaxis] * 0.2
        self.space.set_bounds(abc_bound=abc_bound, **bounds_parameter)

    def set_gpr_params(self, **gpr_params) -> None:
        """Set the parameters of internal GaussianProcessRegressor."""
        self._gpr.set_params(**gpr_params)

    @property
    def space(self):
        """Returns the target space."""
        return self._space

    @property
    def gpr(self):
        """Returns the Gaussian Process regressor."""
        return self._gpr

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(relax_coords={self.relax_coords}, relax_lattice={self.relax_lattice}, "
            f"use_symmetry={self.use_symmetry}"
            f"\n\t\twyckoff_dims={self.wyckoff_dims}, abc_dim={self.abc_dim}, "
            f"\n\t\tangles_dim={self.angles_dim}, kernel={self.gpr.kernel!r}, "
            f"scaler={self.scaler.__class__.__name__}, noisy={self.noisy})"
        )

    def as_dict(self):
        """Dict representation of BayesianOptimizer."""

        def serialize(t) -> tuple:
            """
            Serialize the object.

            Args:
                t: object
            Returns:

            """
            return tuple([int(num) for num in item] if isinstance(item, np.ndarray) else item for item in t)

        def gpr_as_dict(gpr):
            """Serialize the GaussianProcessRegressor."""
            gpr_dict = gpr.get_params()
            gpr_dict["X_train_"] = gpr.X_train_.tolist()
            gpr_dict["alpha_"] = gpr.alpha_.tolist()
            gpr_dict["_y_train_mean"] = gpr._y_train_mean.tolist()
            gpr_dict["_K_inv"] = gpr._K_inv.tolist()
            gpr_kernel = gpr_dict["kernel"]
            kernel_name = gpr_kernel.__class__.__name__
            kernel_params = gpr_kernel.get_params()
            kernel_opt_params = gpr.kernel_.get_params()

            gpr_dict["kernel"] = {"name": kernel_name, "params": kernel_params, "opt_params": kernel_opt_params}

            return gpr_dict

        return {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "structure": self.structure.as_dict(),
            "relax_coords": self.relax_coords,
            "relax_lattice": self.relax_lattice,
            "use_symmetry": self.use_symmetry,
            "space": {
                "params": self.space.params.tolist(),
                "target": self.space.target.tolist(),
                "bounds": self.space.bounds.tolist(),
                "random_state": serialize(self.space.random_state.get_state()),
            },
            "gpr": gpr_as_dict(self.gpr),
            "model": self.model.__class__.__name__,
            "scaler": self.scaler.as_dict(),
            "noisy": self.noisy,
        }

    @classmethod
    def from_dict(cls, d):
        """
        Reconstitute a BayesianOptimizer object from a dict representation of
        BayesianOptimizer created using as_dict().

        Args:
            d (dict): Dict representation of BayesianOptimizer.
        """

        def gpr_from_dict(gpr_d):
            """Instantiate GaussianProcessRegressor from serialization."""
            import sklearn.gaussian_process.kernels as sk_kernels

            d = gpr_d.copy()
            X_train_ = d.pop("X_train_")
            alpha_ = d.pop("alpha_")
            _y_train_mean = d.pop("_y_train_mean")
            _K_inv = d.pop("_K_inv")
            kernel_d = d["kernel"]
            kernel = getattr(sk_kernels, kernel_d["name"])(**kernel_d["params"])
            kernel_ = getattr(sk_kernels, kernel_d["name"])(**kernel_d["opt_params"])
            d["kernel"] = kernel

            gpr = GaussianProcessRegressor()
            gpr.kernel = kernel
            gpr.kernel_ = kernel_
            gpr.set_params(**d)
            gpr.X_train_ = X_train_
            gpr.alpha_ = alpha_
            gpr._y_train_mean = _y_train_mean
            gpr._K_inv = _K_inv

            return gpr

        if d["model"] == "MEGNet":
            import maml.apps.bowsr.model.megnet as energy_model

            model = getattr(energy_model, d["model"])()
        elif d["model"] == "CGCNN":
            import maml.apps.bowsr.model.cgcnn as energy_model

            model = getattr(energy_model, d["model"])()
        elif d["model"] == "DFT":
            import maml.apps.bowsr.model.dft as energy_model

            model = getattr(energy_model, d["model"])()
        else:
            raise AttributeError(f"model {d['model']} is not supported.")

        structure = Structure.from_dict(d["structure"])
        use_symmetry = d["use_symmetry"]
        wps, indices, mapping, lp = struct2perturbation(structure, use_symmetry)
        wyckoff_dims = [wp.dim for wp in wps]
        abc_dim, angles_dim = lp.dims
        relax_coords = d["relax_coords"]
        relax_lattice = d["relax_lattice"]
        noisy = d["noisy"]
        optimizer = BayesianOptimizer(
            model=model,
            structure=structure,
            relax_coords=relax_coords,
            relax_lattice=relax_lattice,
            use_symmetry=use_symmetry,
            noisy=noisy,
        )
        optimizer.structure = structure
        optimizer.wps = wps
        optimizer.indices = indices
        optimizer.mapping = mapping
        optimizer.lp = lp
        optimizer.wyckoff_dims = wyckoff_dims
        optimizer.abc_dim = abc_dim
        optimizer.angles_dim = angles_dim
        gpr = gpr_from_dict(d["gpr"])
        optimizer._gpr = gpr

        space_d = d["space"]
        space_params = np.array(space_d["params"])
        space_target = np.array(space_d["target"])
        space_bounds = np.array(space_d["bounds"])
        space_random_state = np.random.RandomState()  # pylint: disable=E1101
        space_random_state.set_state(space_d["random_state"])
        from maml.apps.bowsr import preprocessing

        scaler = getattr(preprocessing, d["scaler"]["@class"])(**d["scaler"]["params"])
        optimizer.scaler = scaler
        optimizer._space.scaler = scaler
        for x, target in zip(space_params, space_target, strict=False):
            optimizer._space._params = np.concatenate([optimizer._space._params, x.reshape(1, -1)])
            optimizer._space._target = np.concatenate([optimizer._space._target, [target]])
        optimizer._space._bounds = space_bounds
        optimizer._space.random_state = space_random_state

        return optimizer



"""Materials Project DataSource."""

from __future__ import annotations

import pandas as pd
from mp_api.client import MPRester

from maml.base import BaseDataSource


class MaterialsProject(BaseDataSource):
    """Query the Materials Project for Data."""

    def __init__(self, api_key: str | None = None):
        """
        Args:
            api_key (str or None): Materials Project API key.
        """
        self.mpr = MPRester(api_key=api_key)

    def get(self, criteria: str | dict, properties: list[str]) -> pd.DataFrame:  # type: ignore
        """
        Args:
            criteria (str or dict): Criteria for query
            properties (list): Properties to be queried.

        Returns:
            pandas DataFrame
        """
        data = self.mpr.materials.summary.search(**criteria, fields=properties)  # type: ignore[attr-defined]
        data = [{k: v for k, v in dict(d).items() if k in properties} for d in data]
        return pd.DataFrame(data)



"""Symbolic learning."""

from __future__ import annotations

try:
    import cvxpy as cp
except ImportError:
    cp = None

from ._feature_generator import FeatureGenerator, Operator
from ._selectors import SCAD, L0BrutalForce
from ._sis import ISIS, SIS

if cp is None:
    from ._selectors import AdaptiveLasso, DantzigSelector, Lasso
else:
    # import from cvxpy alternatives
    from ._selectors_cvxpy import AdaptiveLassoCP as AdaptiveLasso  # type: ignore
    from ._selectors_cvxpy import DantzigSelectorCP as DantzigSelector  # type: ignore
    from ._selectors_cvxpy import LassoCP as Lasso  # type: ignore

__all__ = [
    "ISIS",
    "SCAD",
    "SIS",
    "AdaptiveLasso",
    "DantzigSelector",
    "FeatureGenerator",
    "L0BrutalForce",
    "Lasso",
    "Operator",
]
